import json

from pybigquery.sqlalchemy_bigquery import BigQueryDialect as BaseDialect


class BigQueryDialect(BaseDialect):
    def create_connect_args(self, url):
        query = url.query

        if 'credentials_info' in query:
            self.credentials_info = json.loads(query.pop('credentials_info'))

            url.query = query

        return super(BigQueryDialect, self).create_connect_args(url)



import json
import numpy
import re

from collections import OrderedDict

from datetime import datetime, timezone

import inflect

from prometheus_client import generate_latest

import pytz

from ..api.appd import AppD
from ..api.extrahop import ExtraHop
from ..database import DatabaseAccessLayer

from ...main import db, aa, aps, set_job_collector_metrics, get_registry, unregister_collector, registered_collectors


def aggregate_values_by_func(func, values):
    if is_number(func):
        value = numpy.percentile(values, int(func))
    else:
        value = 0

        if func == 'avg':
            if values:
                value = numpy.average(values)
        elif func == 'count':
            value = len(values)
        elif func == 'min':
            value = numpy.min(values)
        elif func == 'max':
            value = numpy.max(values)
        elif func == 'sum':
            value = numpy.sum(values)

    return value


def format_label(string):
    label = re.sub(r'[().\'\"]|((?<![/:])/)', '', string)
    label = re.sub(r'^%(?=\w)', 'percent ', label)
    label = re.sub(r'%', ' percent', label)
    label = re.sub(r'>', 'greater than', label)
    label = re.sub(r'<', 'less than', label)
    label = re.sub(r'SQL\*Net', 'Net8', label)
    label = re.sub(r'[ \-]', '_', label)
    label = re.sub(r'\W', '', label)
    label = re.sub(r'_{2,}', '_', label)

    return label.lower()


def format_metric(string):
    metric = re.sub(r'(?<=\S)(/sec)', ' per sec', string)
    metric = re.sub(r'[ \-]', '_', metric)
    metric = re.sub(r'\W', '', metric)
    metric = re.sub(r'_{2,}', '_', metric)

    return metric.lower()


def get_metric_info(service):
    service_name_pattern = re.compile(r'\[(?P<instance>[^\[\]]+)\]', re.X)

    prefix = re.sub(service_name_pattern, '', service)

    instance = None

    m = service_name_pattern.search(service)

    if m is not None:
        components = m.groupdict()

        instance = components['instance']

    return prefix.lower(), instance


def is_element_in_iterable_no_case(element, iterable):
    return element.lower() in [val.lower() for val in list(iterable)]


def is_number(n):
    try:
        float(n)
    except ValueError:
        return False

    return True


def make_label_singular(label):
    p = inflect.engine()

    words = label.split('_')

    words.append(p.singular_noun(words.pop()))

    return '_'.join(words)


def test_aggregation_match(value, aggregation):
    if 'threshold' in aggregation.keys():
        return eval('%d %s %d' % (value, aggregation['threshold']['operator'], int(aggregation['threshold']['value'])))

    return True


def test_aggregation_settings(aggregation, job_name):
    if 'funcs' in aggregation.keys():
        valid_funcs = [
            'avg',
            'count',
            'min',
            'max',
            'sum'
        ]

        for func in aggregation['funcs']:
            if func not in valid_funcs:
                if is_number(func):
                    if int(func) not in range(0, 101):
                        raise ValueError("Invalid aggregation percentile '" + func + "' in job '" + job_name + "'.")
                else:
                    raise ValueError("Unsupported aggregation function '" + func + "' in job '" + job_name + "'.")
    else:
        aggregation['funcs'] = [
            'count'
        ]

    if 'threshold' in aggregation.keys():
        if 'operator' in aggregation['threshold'].keys():
            valid_operator = [
                '>',
                '<',
                '>=',
                '<=',
                '=',
                '<>',
                '!='
            ]

            if aggregation['threshold']['operator'] not in valid_operator:
                raise ValueError("Unsupported aggregation threshold operator '" + aggregation['threshold']['operator'] + "' in job '" + job_name + "'.")
        else:
            raise ValueError("No operator specified for aggregation threshold in job '" + job_name + "'.")

        if 'value' in aggregation['threshold'].keys():
            if is_number(aggregation['threshold']['value']):
                aggregation['threshold']['value'] = aggregation['threshold']['value']
            else:
                raise ValueError("Invalid value '" + aggregation['threshold']['value'] + "' specified for aggregation threshold in job '" + job_name + "'.")
        else:
            raise ValueError("No value specified for aggregation threshold in job '" + job_name + "'.")

    return aggregation


def to_string(n):
    return str(n) if n is not None else ''


class Metrics:
    @staticmethod
    def _get_appdynamics_metrics(job_name, services, application, metric_path, minutes, static_labels=()):
        collector_metrics = {}

        pushgateways = aps.app.config.get('PUSHGATEWAYS')

        all_labels = []

        metric_data = {}

        for service_name in services:
            prefix, instance = get_metric_info(service_name)

            dal = AppD(aa)
            dal.init_aa(service_name)

            tiers = []
            nodes = []

            if application != 'Database Monitoring':
                result = dal.client.get_tiers(application)

                for tier_obj in result:
                    tiers.append(tier_obj.name)

                result = dal.client.get_nodes(application)

                for node_obj in result:
                    nodes.append(node_obj.name)

            options = {
                'time_range_type': 'BEFORE_NOW',
                'duration_in_mins': int(minutes),
                'rollup': True
            }

            result = dal.client.get_metrics(metric_path, application, **options)

            metrics = []

            for metric_obj in result:
                if metric_obj.values:
                    components = metric_obj.path.split('|')

                    if components is not None:
                        metric_profile = [
                            format_label(components.pop(0)),
                            format_label(components.pop())
                        ]

                        label_dict = OrderedDict([
                            ('application', application)
                        ])

                        if len(pushgateways):
                            if instance is not None:
                                label_dict['instance'] = instance

                        if application == 'Database Monitoring':
                            label = make_label_singular(metric_profile[0])

                            label_dict[label] = components.pop(0)

                            if ':' in components[0]:
                                label_dict['node'] = components.pop(0)

                            for component in components:
                                metric_profile.insert(len(metric_profile)-1, format_label(component))
                        else:
                            if metric_profile[0] == 'application_infrastructure_performance':
                                i = 1

                                while i < len(components):
                                    if components[i] != 'Individual Nodes' and components[i] not in nodes:
                                        metric_profile.insert(len(metric_profile)-1, format_label(components[i]))

                                        if components[i-1] == 'JVM':
                                            if components[i] == 'Garbage Collection':
                                                label = make_label_singular(format_label(components.pop(i+1)))

                                                label_dict[label] = components.pop(i+1)
                                            elif components[i] == 'Memory':
                                                label = format_label(components[i])

                                                label_dict[label] = components.pop(i+1)

                                    i += 1
                            else:
                                if metric_profile[0] == 'business_transaction_performance':
                                    label = make_label_singular(format_label(components.pop(0)))

                                    if label == 'business_transaction_group':
                                        label_dict[label] = components.pop(0)
                                    elif label == 'business_transaction':
                                        label_dict[label] = components.pop(1)
                                elif metric_profile[0] in ['errors', 'information_points', 'service_endpoints']:
                                    label = make_label_singular(metric_profile[0])

                                    label_dict[label] = components.pop(1)

                                external_calls_index = 0

                                while len(components) > 0:
                                    if components[0] in tiers:
                                        label_dict['tier'] = components.pop(0)
                                    else:
                                        pattern = re.compile(r'^(?P<desc>Discovered backend call) - (?P<entity>.+)$')

                                        m = pattern.match(components[0])

                                        if m is not None:
                                            subcomponents = m.groupdict()

                                            label = format_label(subcomponents['desc'])

                                            label_dict[label] = subcomponents['entity']

                                            components.pop(0)
                                        else:
                                            label = make_label_singular(format_label(components.pop(0)))

                                            if label == 'external_call':
                                                external_calls_index += 1

                                                label_base = '%s_%d' % (label, external_calls_index)

                                                pattern = re.compile(r'^Call-(?P<type>\S+) to (?P<target>.+)$')

                                                m = pattern.match(components.pop(0))

                                                if m is not None:
                                                    subcomponents = m.groupdict()

                                                    label_dict[label_base] = subcomponents['type']

                                                    pattern = re.compile(r'^(?P<desc>Discovered backend call) - (?P<entity>.+)$')

                                                    m = pattern.match(subcomponents['target'])

                                                    if m is not None:
                                                        subcomponents = m.groupdict()

                                                        label = '%s_%s' % (label_base, format_label(subcomponents['desc']))

                                                        label_dict[label] = subcomponents['entity']
                                                    else:
                                                        label = '%s_tier' % label_base

                                                        label_dict[label] = subcomponents['target']

                                                        if len(components) > 0:
                                                            components.pop(0)
                                            elif label == 'incoming_cross_app_call':
                                                label_base = label

                                                label = '%s_application' % label_base

                                                label_dict[label] = components.pop(0)

                                                pattern = re.compile(r'^Call-(?P<type>\S+) from (?P<source>.+)$')

                                                m = pattern.match(components.pop(0))

                                                if m is not None:
                                                    subcomponents = m.groupdict()

                                                    label_dict[label_base] = subcomponents['type']

                                                    label = '%s_tier' % label_base

                                                    label_dict[label] = subcomponents['source']
                                            elif label in ['individual_node', 'thread_task']:
                                                label_dict[label] = components.pop(0)

                        row = metric_obj.values[0].__dict__

                        metric_dict = OrderedDict([
                            (key, value) for key, value in row.items() if key != 'start_time_ms'
                        ])

                        if not metrics:
                            metrics = [
                                metric for metric, value in metric_dict.items() if is_number(value)
                            ]

                        label_dict.update(OrderedDict([
                            (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                        ]))

                        for label in label_dict.keys():
                            if label not in all_labels:
                                all_labels.append(label)

                        timestamp = row['start_time_ms'] / 1000 + (int(minutes) * 60)

                        for metric in metrics:
                            metric_name = '%s_%s_%s' % (prefix, '_'.join(metric_profile), metric)

                            if metric_name not in metric_data.keys():
                                metric_data[metric_name] = []

                            metric_data[metric_name].append((
                                label_dict,
                                int(metric_dict[metric]),
                                timestamp
                            ))

        for metric_name, data in metric_data.items():
            for (label_dict, value, timestamp) in data:
                normalized_label_dict = OrderedDict()

                for label in all_labels:
                    if label in label_dict.keys():
                        normalized_label_dict[label] = label_dict[label]
                    else:
                        normalized_label_dict[label] = ''

                json_label_data = json.dumps(normalized_label_dict)

                if metric_name not in collector_metrics.keys():
                    collector_metrics[metric_name] = {}

                if json_label_data not in collector_metrics[metric_name].keys():
                    collector_metrics[metric_name][json_label_data] = (value, timestamp)

        return collector_metrics

    @staticmethod
    def _get_database_metrics(job_name, services, statement, value_columns, static_labels=(), timestamp_column=None, timezones={}):
        def is_value_column(column):
            return is_element_in_iterable_no_case(column, value_columns)

        collector_metrics = {}

        pushgateways = aps.app.config.get('PUSHGATEWAYS')

        for service_name in services:
            prefix, instance = get_metric_info(service_name)

            dal = DatabaseAccessLayer(db)
            dal.init_db(service_name)

            result = dal.execute(statement)

            timestamp = datetime.now(timezone.utc).timestamp()

            db_tzinfo = None

            if timestamp_column is not None and service_name in timezones.keys():
                db_tzinfo = pytz.timezone(timezones[service_name])

            normalized_value_columns = None
            normalized_timestamp_column = None

            for row in result:
                def is_unmatched_value_column(column):
                    return not is_element_in_iterable_no_case(column, row.keys())

                if normalized_value_columns is None:
                    normalized_value_columns = list(filter(is_value_column, row.keys()))

                    unmatched_value_columns = list(filter(is_unmatched_value_column, value_columns))

                    if unmatched_value_columns:
                        raise ValueError("Value column(s) " + ", ".join(["'" + column + "'" for column in unmatched_value_columns]) + " specified in job '" + job_name + "' not returned in query result.")

                if timestamp_column is not None and normalized_timestamp_column is None:
                    for column in row.keys():
                        if timestamp_column.lower() == column.lower():
                            normalized_timestamp_column = column

                    if normalized_timestamp_column is None:
                        raise ValueError("Timestamp column '" + timestamp_column + "' specified in job '" + job_name + "' not returned in query result.")

                label_dict = OrderedDict()

                if len(pushgateways):
                    if instance is not None:
                        label_dict['instance'] = instance

                label_dict.update(OrderedDict([
                    (format_label(column), to_string(row[column])) for column in row.keys() if column not in normalized_value_columns and column != normalized_timestamp_column
                ]))

                label_dict.update(OrderedDict([
                    (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                ]))

                if db_tzinfo is not None:
                    if isinstance(row[normalized_timestamp_column], datetime):
                        if row[normalized_timestamp_column].tzinfo is not None and row[normalized_timestamp_column].tzinfo.utcoffset(row[normalized_timestamp_column]) is not None:
                            timestamp = row[normalized_timestamp_column].timestamp()
                        else:
                            timestamp = row[normalized_timestamp_column].astimezone(db_tzinfo).timestamp()

                json_label_data = json.dumps(label_dict)

                for column in normalized_value_columns:
                    metric_name = '%s_%s' % (prefix, format_metric(column))

                    if metric_name not in collector_metrics.keys():
                        collector_metrics[metric_name] = {}

                    if json_label_data not in collector_metrics[metric_name].keys():
                        collector_metrics[metric_name][json_label_data] = (row[column], timestamp)

        return collector_metrics

    @staticmethod
    def _get_extrahop_metrics(job_name, services, params, metric, aggregation, minutes, static_labels=()):
        collector_metrics = {}

        aggregation = test_aggregation_settings(aggregation, job_name)

        pushgateways = aps.app.config.get('PUSHGATEWAYS')

        for service_name in services:
            prefix, instance = get_metric_info(service_name)

            dal = ExtraHop(aa)
            dal.init_aa(service_name)

            options = {**params, **{
                'from': '-%sm' % minutes,
                'until': 0
            }}

            result = dal.client.get_metrics(**options)

            timestamp = datetime.now(timezone.utc).timestamp()

            if 'stats' in result.keys():
                metric_dict = OrderedDict()

                for row in result['stats']:
                    if row['values']:
                        if row['values'][0]:
                            metric_spec_name = row['values'][0][0]['key']['str']

                            if metric_spec_name not in metric_dict.keys():
                                metric_dict[metric_spec_name] = []

                            value = int(row['values'][0][0]['value'])

                            if test_aggregation_match(value, aggregation):
                                metric_dict[metric_spec_name].append(value)

                for metric_spec_name, values in metric_dict.items():
                    label_dict = OrderedDict([
                        ('metric_spec_name', metric_spec_name.lower())
                    ])

                    if len(pushgateways):
                        if instance is not None:
                            label_dict['instance'] = instance

                    label_dict.update(OrderedDict([
                        (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                    ]))

                    json_label_data = json.dumps(label_dict)

                    for func in aggregation['funcs']:
                        metric_name = '%s_%s_%s' % (prefix, format_metric(metric), func.lower())

                        if metric_name not in collector_metrics.keys():
                            collector_metrics[metric_name] = {}

                        if json_label_data not in collector_metrics[metric_name].keys():
                            collector_metrics[metric_name][json_label_data] = (aggregate_values_by_func(func, values), timestamp)

        return collector_metrics

    @staticmethod
    def generate_metrics(*args):
        if len(args) >= 2:
            category = args[0]
            job_name = args[1]

            func_name = '_get_' + category + '_metrics'

            if hasattr(Metrics, func_name):
                with aps.app.app_context():
                    try:
                        aps.app.logger.info("Job '" + job_name + "' started.")

                        func = getattr(Metrics, func_name)
                        collector_metrics = func(*args[1:])

                        set_job_collector_metrics(job_name, collector_metrics)

                        aps.app.logger.info("Job '" + job_name + "' completed.")
                    except Exception as e:
                        aps.app.logger.warning("Job '" + job_name + "' failed. " + type(e).__name__ + ": " + str(e))

                        if aps.app.config.get('SUSPEND_JOB_ON_FAILURE'):
                            aps.pause_job(job_name)

                            aps.app.logger.warning("Job '" + job_name + "' suspended.")

                        if job_name in registered_collectors.keys():
                            unregister_collector(job_name, registered_collectors[job_name])

                            del registered_collectors[job_name]

    @staticmethod
    def read_prometheus_metrics(job_name):
        registry = get_registry(job_name)

        return generate_latest(registry)



import json
import logging

from datetime import datetime, timedelta

from flask import Flask
from flask_apialchemy import APIAlchemy
from flask_apscheduler import APScheduler
from flask_sqlalchemy import _EngineConnector as _EngineConnectorBase, SQLAlchemy as SQLAlchemyBase

from prometheus_client import generate_latest
from prometheus_client.core import GaugeMetricFamily

from prometheus_flask_exporter import PrometheusMetrics

from sqlalchemy.dialects import registry

from werkzeug.middleware.proxy_fix import ProxyFix

from .config import config_by_name

from .util.prometheus_helper import prometheus_multiproc_dir, get_registry, register_collector, unregister_collector


class _EngineConnector(_EngineConnectorBase):
    def get_options(self, sa_url, echo):
        sa_url, options = super(_EngineConnector, self).get_options(sa_url, echo)

        if sa_url.drivername.startswith('oracle'):
            # options.setdefault('max_identifier_length', 30)

            options.setdefault('connect_args', {
                'events': True
            })

        poolclass = options.get('poolclass')

        if poolclass and poolclass.__name__ in ['NullPool', 'StaticPool']:
            if 'pool_size' in options.keys():
                options.pop('pool_size')

            if 'pool_recycle' in options.keys():
                options.pop('pool_recycle')

        return sa_url, options


class SQLAlchemy(SQLAlchemyBase):
    def make_connector(self, app=None, bind=None):
        """Creates the connector for a given state and bind."""
        return _EngineConnector(self, self.get_app(app), bind)


db = SQLAlchemy()
aa = APIAlchemy()

registry.register('bigquery', __package__ + '.database.bigquery.dialect', 'BigQueryDialect')
registry.register('bigquery.pybigquery', __package__ + '.database.bigquery.dialect', 'BigQueryDialect')
registry.register('db2', __package__ + '.database.db2.dialect', 'DB2Dialect')
registry.register('db2.ibm_db', __package__ + '.database.db2.dialect', 'DB2Dialect')

aps = APScheduler()

app_registry_name = 'application'

metrics = PrometheusMetrics.for_app_factory(registry=get_registry(app_registry_name))

registered_collectors = {}

job_collector_metrics = {}


def create_app(config_name):
    from .. import blueprint

    app = Flask(__name__)

    app.wsgi_app = ProxyFix(app.wsgi_app)

    try:
        config_obj = config_by_name[config_name]()

        app.config.from_object(config_obj)

        if not app.debug:
            logger = logging.getLogger('werkzeug')
            logger.setLevel(logging.ERROR)

            # app.logger.setLevel(logging.ERROR)

        db.init_app(app)
        aa.init_app(app)

        with app.app_context():
            config_obj.add_jobs_from_source(aa)

        app.config.from_object(config_obj)

        aps.init_app(app)
        metrics.init_app(app)

        app.register_blueprint(blueprint)
    except ValueError as error:
        app.logger.critical(error)
        exit(1)

    return app


def delete_job(job_name, pushgateways):
    if job_name in registered_collectors.keys():
        if len(pushgateways):
            options = {
                'grouping_key': {
                    'job': job_name
                }
            }

            if registered_collectors[job_name].instance:
                options['grouping_key']['instance'] = registered_collectors[job_name].instance

            for service, pushgateway in pushgateways.items():
                try:
                    pushgateway.delete(job=job_name, **options)
                except Exception as e:
                    aps.app.logger.warning("Failed removing metrics for job '" + job_name + "' from Pushgateway service '" + service + "'. " + type(e).__name__ + ": " + str(e))

        unregister_collector(job_name, registered_collectors[job_name])

        del registered_collectors[job_name]


def get_jobs():
    job_list = aps.app.config.get('SCHEDULER_JOBS')

    return [
        job['id'] for job in job_list
    ]


def init_scheduler(config_name):
    try:
        job_list = aps.app.config.get('SCHEDULER_JOBS')

        config_obj = config_by_name[config_name]()

        with aps.app.app_context():
            config_obj.set_pushgateways(aa)

        aps.app.config.from_object(config_obj)

        aps.app.config['SCHEDULER_JOBS'] = job_list

        source_refresh_interval = aps.app.config.get('JOBS_SOURCE_REFRESH_INTERVAL')

        if source_refresh_interval is not None:
            source_refresh_job_name = 'SOURCE_REFRESH'

            job_kwargs = {
                'trigger': 'interval',
                'args': [source_refresh_job_name, config_name],
                'seconds': int(source_refresh_interval) * 60
            }

            aps.add_job(source_refresh_job_name, __package__ + ':update_jobs', **job_kwargs)
    except ValueError as error:
        aps.app.logger.critical(error)
        exit(1)


def set_job_collector_metrics(job_name, collector_metrics):
    job_collector_metrics[job_name] = collector_metrics

    if job_name not in registered_collectors:
        registered_collectors[job_name] = JobCollector(job_name)

        register_collector(job_name, registered_collectors[job_name])

    pushgateways = aps.app.config.get('PUSHGATEWAYS')

    if len(pushgateways):
        job_collector_registry = get_registry(job_name)

        if registered_collectors[job_name].instance is None:
            generate_latest(job_collector_registry)

        options = {
            'grouping_key': {
                'job': job_name
            }
        }

        if registered_collectors[job_name].instance:
            options['grouping_key']['instance'] = registered_collectors[job_name].instance

        for service, pushgateway in pushgateways.items():
            try:
                pushgateway.push(job=job_name, registry=job_collector_registry, **options)

                aps.app.logger.info("Sent metrics for job '" + job_name + "' to Pushgateway service '" + service + "'.")
            except Exception as e:
                aps.app.logger.warning("Failed sending metrics for job '" + job_name + "' to Pushgateway service '" + service + "'. " + type(e).__name__ + ": " + str(e))


def shutdown_scheduler():
    aps.shutdown()


def start_scheduler(run_jobs=True):
    aps.start()

    initial_run_time = datetime.now()

    for job_name in get_jobs():
        if run_jobs:
            initial_run_time += timedelta(seconds=aps.app.config.get('SCHEDULER_JOB_DELAY_SECONDS'))

            aps.modify_job(job_name, next_run_time=initial_run_time)
        else:
            aps.pause_job(job_name)


def update_jobs(source_refresh_job_name, config_name):
    try:
        aps.app.logger.info("Check for job updates started.")

        config_obj = config_by_name[config_name]()

        with aps.app.app_context():
            config_obj.set_pushgateways(aa)

            config_obj.add_jobs_from_source(aa)

        job_list = aps.app.config.get('SCHEDULER_JOBS')

        pushgateways = config_obj.PUSHGATEWAYS

        jobs = {}

        for i in range(len(job_list)):
            job_name = job_list[i]['id']

            if job_name not in config_obj.jobs.keys():
                delete_job(job_name, pushgateways)

                aps.remove_job(job_name)

                aps.app.logger.info("Job '" + job_name + "' removed.")
            else:
                jobs[job_name] = job_list[i]

        for i in range(len(config_obj.SCHEDULER_JOBS)):
            job_name = config_obj.SCHEDULER_JOBS[i]['id']

            run_job = True

            if job_name in jobs.keys():
                run_job = json.dumps(config_obj.SCHEDULER_JOBS[i]) != json.dumps(jobs[job_name])

                if run_job:
                    delete_job(job_name, pushgateways)

                    aps.modify_job(**config_obj.SCHEDULER_JOBS[i])

                    aps.app.logger.info("Job '" + job_name + "' updated.")
            else:
                aps.add_job(**config_obj.SCHEDULER_JOBS[i])

                aps.app.logger.info("Job '" + job_name + "' added.")

            if run_job:
                aps.run_job(job_name)

        aps.app.config.from_object(config_obj)

        aps.app.logger.info("Check for job updates completed.")
    except Exception as e:
        aps.app.logger.warning("Check for job updates failed. " + type(e).__name__ + ": " + str(e))

        aps.pause_job(source_refresh_job_name)


class JobCollector:
    instance = None

    def __init__(self, job_name):
        self._job_name = job_name

    def collect(self):
        pushgateways = aps.app.config.get('PUSHGATEWAYS')

        for metric_name, info in job_collector_metrics[self._job_name].items():
            json_label_data = list(info.keys())[0]

            label_dict = json.loads(json_label_data)

            g = GaugeMetricFamily(metric_name, '', labels=label_dict.keys())

            if self.instance is None:
                if 'instance' in label_dict.keys():
                    self.instance = label_dict.pop('instance')
                else:
                    self.instance = ''

            for json_label_data, metric in info.items():
                label_dict = json.loads(json_label_data)

                options = {}

                if len(pushgateways) == 0:
                    options['timestamp'] = metric[1]

                g.add_metric(label_dict.values(), metric[0], **options)

            yield g



from flask_restx import Resource

from flask_apscheduler import api as aps_api

from ..util.dto import SchedulerDto

api = SchedulerDto.api


@api.route('')
class Scheduler(Resource):
    @api.doc('get_scheduler_info')
    def get(self):
        """Gets the scheduler info."""
        return aps_api.get_scheduler_info()


@api.route('/jobs')
class SchedulerJobList(Resource):
    @api.doc('get_jobs')
    def get(self):
        """Gets all scheduled jobs."""
        return aps_api.get_jobs()

    # @api.doc('add_job')
    # def post(self):
    #     """Adds a new job."""
    #     return aps_api.add_job()


@api.route('/jobs/<job_id>')
class SchedulerJob(Resource):
    @api.doc('get_job')
    def get(self, job_id):
        """Gets a job."""
        return aps_api.get_job(job_id)

    # @api.doc('delete_job')
    # def delete(self, job_id):
    #     """Deletes a job."""
    #     return aps_api.delete_job(job_id)

    # @api.doc('update_job')
    # def patch(self, job_id):
    #     """Updates a job."""
    #     return aps_api.update_job(job_id)


@api.route('/jobs/<job_id>/pause')
class SchedulerJobPause(Resource):
    @api.doc('pause_job')
    def get(self, job_id):
        """Pauses a job."""
        return aps_api.pause_job(job_id)


@api.route('/jobs/<job_id>/resume')
class SchedulerJobResume(Resource):
    @api.doc('resume_job')
    def get(self, job_id):
        """Resumes a job."""
        return aps_api.resume_job(job_id)


@api.route('/jobs/<job_id>/run')
class SchedulerJobRun(Resource):
    @api.doc('run_job')
    def get(self, job_id):
        """Runs a job."""
        return aps_api.run_job(job_id)



from sqlalchemy import text


class DatabaseAccessLayer:
    _session = None

    def __init__(self, db):
        self._db = db

    def execute(self, statement):
        result = self._session.execute(text(statement))

        return result

    def init_db(self, bind):
        engine = self._db.get_engine(bind=bind)

        options = {}

        backend = engine.url.get_backend_name()

        if backend == 'informix':
            options['isolation_level'] = 'DIRTY READ'

        self._session = self._db.create_scoped_session({
            'autocommit': False,
            'autoflush': False,
            'bind': engine.execution_options(**options)
        })



import os
import re
import yaml

import pytz

package_components = __package__.split('.')

package_components.pop()

job_func_root = '.'.join(package_components)


def apply_job_templates(jobs, templates):
    for job_name, credentials in jobs.items():
        if 'use_template' in credentials.keys():
            template_name = credentials.pop('use_template')

            if template_name in templates.keys():
                template = templates[template_name].copy()

                template.update(credentials)

                jobs[job_name].update(template)
            else:
                raise ValueError("Job '" + job_name + "' references undefined template: " + template_name + ".")

    return jobs


def build_job(category, name, bind, seconds, *args):
    return {
        'func': job_func_root + '.service.metrics_service:Metrics.generate_metrics',
        'trigger': 'interval',
        'args': (category, name, bind) + args,
        'id': name,
        'seconds': seconds
    }


def build_job_list(jobs, apialchemy_info, sqlalchemy_info):
    """Builds list of scheduled jobs to assign to SCHEDULER_JOBS"""
    job_list = []

    db_default_local_tz = os.getenv('DB_DEFAULT_LOCAL_TZ')

    if db_default_local_tz not in pytz.all_timezones:
        db_default_local_tz = 'UTC'

    apialchemy_prefix, apialchemy_binds = apialchemy_info
    sqlalchemy_prefix, sqlalchemy_binds = sqlalchemy_info

    service_name_pattern = re.compile(r'^' + r'(?:(?:' + re.escape(apialchemy_prefix) + r')|(?:' + re.escape(sqlalchemy_prefix) + r'))(?P<name>.+)$', re.X)

    for job_name, credentials in jobs.items():
        if 'services' in credentials.keys() and 'interval_minutes' in credentials.keys():
            service_names = {}

            for service in credentials['services']:
                m = service_name_pattern.match(service)

                if m is not None:
                    components = m.groupdict()

                    service_names[service] = components['name']
                else:
                    raise ValueError("Invalid service name: '" + service + "'.")

            for k, v in credentials.items():
                if isinstance(v, str):
                    undefined_env_vars, credentials[k] = populate_env_vars(v)

                    if undefined_env_vars:
                        raise ValueError("Job '" + job_name + "' references undefined environment variable(s): " + ", ".join(undefined_env_vars) + ".")

            valid_service_names = list(apialchemy_binds.keys()) + list(sqlalchemy_binds.keys())

            for service, name in service_names.items():
                if name not in valid_service_names:
                    raise ValueError("Service '" + service + "' not found for job '" + job_name + "'.")

            if all(service_name in apialchemy_binds.keys() and 'vendor' in apialchemy_binds[service_name].keys() for service_name in service_names.values()):
                vendor = None

                for service_name in service_names.values():
                    if vendor is None:
                        vendor = apialchemy_binds[service_name]['vendor']

                    if apialchemy_binds[service_name]['vendor'] != vendor:
                        raise ValueError("Services defined for job '" + job_name + "' must reference same API vendor.")

                job_args = [
                    vendor,
                    job_name,
                    list(service_names.values()),
                    int(credentials['interval_minutes']) * 60
                ]

                service_job_args = None

                if vendor == 'appdynamics':
                    if 'application' in credentials.keys():
                        job_args.append(credentials['application'])

                        if 'metric_path' in credentials.keys():
                            service_job_args = [
                                credentials['metric_path'],
                                credentials['interval_minutes']
                            ]
                elif vendor == 'extrahop':
                    if 'metric_params' in credentials.keys() and 'metric_name' in credentials.keys() and 'aggregation' in credentials.keys():
                        service_job_args = [
                            credentials['metric_params'],
                            credentials['metric_name'],
                            credentials['aggregation'],
                            credentials['interval_minutes']
                        ]

                if service_job_args is not None:
                    job_args += service_job_args

                    if 'static_labels' in credentials.keys():
                        static_label_list = list(filter(None, re.split(r'\s*,\s*', credentials['static_labels'])))

                        job_args.append([
                            tuple(filter(None, label_str.split(':'))) for label_str in static_label_list
                        ])

                    job_list.append(build_job(*job_args))

                    continue
            elif all(service_name in sqlalchemy_binds.keys() for service_name in service_names.values()):
                if 'statement' in credentials.keys() and 'value_columns' in credentials.keys():
                    value_columns = list(filter(None, re.split(r'\s*,\s*', credentials['value_columns'])))

                    if not value_columns:
                        raise ValueError("No value columns specified for job '" + job_name + "'.")

                    job_args = [
                        'database',
                        job_name,
                        list(service_names.values()),
                        int(credentials['interval_minutes']) * 60,
                        credentials['statement'],
                        value_columns
                    ]

                    if 'static_labels' in credentials.keys():
                        static_label_list = list(filter(None, re.split(r'\s*,\s*', credentials['static_labels'])))

                        job_args.append([
                            tuple(filter(None, label_str.split(':'))) for label_str in static_label_list
                        ])

                    if 'timestamp_column' in credentials.keys():
                        job_args.append(credentials['timestamp_column'])

                        timezones = {service_name: sqlalchemy_binds[service_name]['timezone'] if 'timezone' in sqlalchemy_binds[service_name].keys() else db_default_local_tz for service_name in service_names.values()}

                        job_args.append(timezones)

                    job_list.append(build_job(*job_args))

                    continue
            else:
                raise ValueError("Services defined for job '" + job_name + "' must reference same connection type.")

        raise ValueError("Missing required credential(s) for job '" + job_name + "'.")

    return job_list


def get_jobs_from_services(job_prefix, template_prefix, services):
    jobs = {}
    templates = {}

    job_name_pattern = re.compile(r'^' + re.escape(job_prefix) + r'(?P<name>.+)$', re.X)
    template_name_pattern = re.compile(r'^' + re.escape(template_prefix) + r'.+$', re.X)

    for service in services:
        m = job_name_pattern.match(service.name)

        if m is not None:
            components = m.groupdict()

            name = components['name']

            jobs[name] = service.credentials
        else:
            m = template_name_pattern.match(service.name)

            if m is not None:
                templates[service.name] = service.credentials

    return apply_job_templates(jobs, templates)


def get_jobs_from_source(aa, apialchemy_info):
    jobs = {}

    apialchemy_prefix, apialchemy_binds = apialchemy_info

    service_name_pattern = re.compile(r'^' + r'(?:' + re.escape(apialchemy_prefix) + r')(?P<name>.+)$', re.X)

    api_vendor_pattern = re.compile(r'^(?:(?P<vendor>\w+)(?:\+(?:http|https))?)(?=://)', re.X)

    jobs_source_service = os.getenv('JOBS_SOURCE_SERVICE')

    if jobs_source_service is not None:
        m = service_name_pattern.match(jobs_source_service)

        if m is not None:
            components = m.groupdict()

            service_name = components['name']

            if service_name in apialchemy_binds.keys():
                conn_str = apialchemy_binds[service_name]

                m = api_vendor_pattern.match(conn_str)

                if m is not None:
                    components = m.groupdict()

                    if components['vendor'] == 'github':
                        from ..api.github import GitHub

                        jobs_source_org = os.getenv('JOBS_SOURCE_ORG')
                        jobs_source_repo = os.getenv('JOBS_SOURCE_REPO')

                        if jobs_source_org is not None and jobs_source_repo is not None:
                            jobs_source_path = os.getenv('JOBS_SOURCE_PATH')

                            if jobs_source_path is not None:
                                jobs_source_branch = os.getenv('JOBS_SOURCE_BRANCH', 'master')

                                dal = GitHub(aa)
                                dal.init_aa(service_name)

                                contents = dal.service.get_file_contents(
                                    dal.client,
                                    jobs_source_org,
                                    jobs_source_repo,
                                    jobs_source_path,
                                    jobs_source_branch
                                )

                                jobs = get_services_from_yaml(yaml.safe_load(contents))

                                jobs_source_templates_path = os.getenv('JOBS_SOURCE_TEMPLATES_PATH')

                                if jobs_source_templates_path is not None:
                                    contents = dal.service.get_file_contents(
                                        dal.client,
                                        jobs_source_org,
                                        jobs_source_repo,
                                        jobs_source_templates_path,
                                        jobs_source_branch
                                    )

                                    templates = get_services_from_yaml(yaml.safe_load(contents))

                                    jobs = apply_job_templates(jobs, templates)
                        else:
                            raise ValueError("Missing environment variable(s) required for jobs source service '" + jobs_source_service + "'.")
            else:
                raise ValueError("Service '" + jobs_source_service + "' not found.")

    return jobs


def get_services_from_yaml(yaml_data):
    services = {}

    if isinstance(yaml_data, dict):
        for service, credentials in yaml_data.items():
            if isinstance(credentials, dict):
                services[service] = credentials

    return services


def populate_env_vars(string):
    undefined_env_vars = []

    env_var_pattern = re.compile(r'%%(?P<var>\w+)%%', re.X)

    for m in env_var_pattern.finditer(string):
        components = m.groupdict()

        env_var = os.getenv(components['var'])

        if env_var is None:
            undefined_env_vars.append(components['var'])
        else:
            current_env_var_pattern = re.compile(r'%%' + re.escape(components['var']) + r'%%', re.X)

            string = current_env_var_pattern.sub(env_var, string, 1)

    return undefined_env_vars, string



import os
import re

from prometheus_client.core import CollectorRegistry
from prometheus_client.multiprocess import MultiProcessCollector

collector_registries = {}

prometheus_multiproc_dir = os.getenv('prometheus_multiproc_dir')


def get_pushgateways(aa, apialchemy_info):
    pushgateways = {}

    apialchemy_prefix, apialchemy_binds = apialchemy_info

    service_name_pattern = re.compile(r'^' + r'(?:' + re.escape(apialchemy_prefix) + r')(?P<name>.+)$', re.X)

    api_vendor_pattern = re.compile(r'^(?:(?P<vendor>\w+)(?:\+(?:http|https))?)(?=://)', re.X)

    pushgateway_services = list(filter(None, re.split(r'\s*,\s*', os.getenv('PUSHGATEWAY_SERVICES', ''))))

    for service in pushgateway_services:
        m = service_name_pattern.match(service)

        if m is not None:
            components = m.groupdict()

            service_name = components['name']

            if service_name in apialchemy_binds.keys():
                conn_str = apialchemy_binds[service_name]

                m = api_vendor_pattern.match(conn_str)

                if m is not None:
                    components = m.groupdict()

                    if components['vendor'] == 'pushgateway':
                        from ..api.pushgateway import Pushgateway

                        dal = Pushgateway(aa)
                        dal.init_aa(service_name)

                        pushgateways[service] = dal.client
                    else:
                        raise ValueError("Service '" + service + "' is not a valid Pushgateway.")
            else:
                raise ValueError("Service '" + service + "' not found.")

    return pushgateways


def get_registry(name):
    if name not in collector_registries.keys():
        collector_registries[name] = CollectorRegistry()

        if prometheus_multiproc_dir is not None:
            MultiProcessCollector(collector_registries[name])

    return collector_registries[name]


def register_collector(name, collector):
    job_registry = get_registry(name)

    job_registry.register(collector)


def unregister_collector(name, collector):
    if name in collector_registries.keys():
        collector_registries[name].unregister(collector)

        del collector_registries[name]



from abc import ABCMeta, abstractmethod


class ApiAccessLayer:
    __metaclass__ = ABCMeta

    service = None

    def __init__(self, aa):
        self._aa = aa

    @abstractmethod
    def init_aa(self, bind):
        self.service = self._aa.get_service(bind=bind)



from cryptofy import encoding, decrypt

import re

from urllib import parse

from .misc_helper import str_to_bool


def build_conn_str(credentials):
    """Constructs API connection string from credentials."""
    base = []

    if 'apikey' in credentials:
        base.append(parse.quote(credentials['apikey']))
    else:
        if credentials['vendor'] == 'appdynamics':
            username_account = []

            if 'username' in credentials:
                username_account.append(parse.quote(credentials['username']))

                if 'account' in credentials:
                    username_account.append(parse.quote(credentials['account']))

            if len(username_account):
                username_account_password = [
                    '@'.join(username_account)
                ]

                if 'password' in credentials:
                    username_account_password.append(parse.quote(credentials['password']))

                base.append(':'.join(username_account_password))
        else:
            username_password = []

            if 'username' in credentials:
                username_password.append(parse.quote(credentials['username']))

                if 'password' in credentials:
                    username_password.append(parse.quote(credentials['password']))

            if len(username_password):
                base.append(':'.join(username_password))

    hostname_port = []

    scheme = [credentials['vendor']]

    if 'hostname' in credentials:
        url_components = parse.urlparse(credentials['hostname'])

        if url_components.scheme:
            scheme.append(url_components.scheme)

        if url_components.hostname is not None:
            hostname = url_components.hostname
        else:
            hostname = credentials['hostname']

        hostname_port.append(hostname)

    if 'port' in credentials:
        hostname_port.append(str(credentials['port']))

    if len(hostname_port):
        base.append(':'.join(hostname_port))

    uri = ['+'.join(scheme)]

    if len(base):
        uri.append('@'.join(base))

    return '://'.join(uri)


def build_bind_dict(binds, key=''):
    """Builds dict of connections to assign to APIALCHEMY_BINDS"""
    bind_dict = {}

    encrypted_credentials = [
        'apikey',
        'password'
    ]

    for name, credentials in binds.items():
        if 'encrypted' in credentials.keys():
            if str_to_bool(credentials['encrypted']):
                if key != '':
                    for i in encrypted_credentials:
                        if i in credentials.keys():
                            credentials[i] = decrypt(bytes(key, encoding=encoding), credentials[i]).decode(encoding)
                else:
                    raise ValueError('No secret key found.')

        bind_dict[name] = build_conn_str(credentials)

    return bind_dict


def parse_services_for_binds(prefix, services):
    binds = {}

    service_name_pattern = re.compile(r'^' + re.escape(prefix) + r'(?P<name>.+)$', re.X)

    for service in services:
        m = service_name_pattern.match(service.name)

        if m is not None:
            components = m.groupdict()

            name = components['name']

            binds[name] = service.credentials

    return binds



import os
import re

from setuptools import find_packages, setup

from discover_tests import DiscoverTest


def get_requirements_from_file(extra=None):
    requirements = []

    filename = 'requirements.txt'

    if extra is not None:
        filename = extra + '-' + filename

    with open(
        os.path.join(os.path.dirname(__file__), 'pip', filename)
    ) as file:
        lines = file.readlines()

        specifier_pattern = re.compile(r"""^(?!-|(?:\s*$))([^\n]+)""", re.S)

        for line in lines:
            m = specifier_pattern.match(line)

            if m is not None:
                requirements.append(m.group(1))

    return requirements


def str_to_bool(x):
    x = str(x)

    return x.lower() in ('true', 't', 'yes', 'y', '1')


with open(
    os.path.join(os.path.dirname(__file__), 'metREx', 'app', '__init__.py')
) as i_file:
    info = i_file.read()

    TITLE = (
        re.compile(r""".*__title__ = ["']([^\n]*)['"]""", re.S).match(info).group(1)
    )

    VERSION = (
        re.compile(r""".*__version__ = ["']([^\n]*)['"]""", re.S).match(info).group(1)
    )

    DESCRIPTION = (
        re.compile(r""".*__description__ = ["']([^\n]*)['"]""", re.S).match(info).group(1)
    )

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as r_file:
    README = r_file.read()

INSTALL_REQUIRES = list(get_requirements_from_file())

EXTRAS = [
    'bigquery',
    'db2',
    'informix',
    'mssql',
    'mysql',
    'oracle',
    'postgresql',
    'sqlite'
]

EXTRAS_REQUIRE = {extra: list(get_requirements_from_file(extra)) for extra in EXTRAS}

install_all_extras = str_to_bool(os.getenv('INSTALL_ALL_EXTRAS', False))

if install_all_extras:
    for specifiers in EXTRAS_REQUIRE.values():
        for specifier in specifiers:
            if specifier not in INSTALL_REQUIRES:
                INSTALL_REQUIRES.append(specifier)

TESTS_REQUIRE = list(get_requirements_from_file('test'))

setup(name=TITLE,
      version=VERSION,
      description=DESCRIPTION,
      long_description=README,
      long_description_content_type='text/markdown',
      license='Apache 2.0',
      keywords='prometheus exporter prometheus-exporter metrics metrics-exporter query-exporter sql sql-exporter appd appd-exporter appdynamics appdynamics-exporter extrahop extrahop-exporter',
      url='https://github.com/homedepot/metREx',
      author='Mike Phillipson',
      author_email='MICHAEL_PHILLIPSON1@homedepot.com',
      packages=find_packages(include=[
          'metREx',
          'metREx.*'
      ]),
      data_files=[
          ('static', [
              'favicon.ico'
          ])
      ],
      entry_points={
          'console_scripts': [
              'metrex=metREx.manage:run'
          ]
      },
      python_requires='>=3.6',
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      tests_require=TESTS_REQUIRE,
      cmdclass={
          'test': DiscoverTest
      },
      zip_safe=False)
